import logging
import os
import re
from pathlib import Path

from cg.apps.demultiplex.sample_sheet.read_sample_sheet import get_sample_sheet_from_file
from cg.apps.demultiplex.sample_sheet.sample_sheet_models import SampleSheet
from cg.constants.constants import FileExtensions
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.sequencing import FLOWCELL_Q30_THRESHOLD, Sequencers
from cg.io.csv import read_csv, write_csv
from cg.models.flow_cell.flow_cell import FlowCellDirectoryData
from cg.utils.files import (
    get_file_in_directory,
    get_files_matching_pattern,
    is_pattern_in_file_path_name,
    rename_file,
)

LOG = logging.getLogger(__name__)

NANOPORE_SEQUENCING_SUMMARY_PATTERN: str = r"final_summary_*.txt"


def get_lane_from_sample_fastq(sample_fastq_path: Path) -> int:
    """
    Extract the lane number from the sample fastq path.
    Pre-condition:
        - The fastq file name contains the lane number formatted as _L<lane_number>
    """
    pattern = r"_L(\d+)"
    lane_match = re.search(pattern, sample_fastq_path.name)

    if lane_match:
        return int(lane_match.group(1))

    raise ValueError(f"Could not extract lane number from fastq file name {sample_fastq_path.name}")


def is_sample_id_in_directory_name(directory: Path, sample_internal_id: str) -> bool:
    """Validate that directory name is formatted as Sample_<sample_id> or Sample_<sample_id>_."""
    sample_pattern: str = f"Sample_{sample_internal_id}"
    return f"{sample_pattern}_" in directory.name or sample_pattern == directory.name


def is_sample_id_in_file_name(sample_fastq: Path, sample_internal_id: str) -> bool:
    """Validate that file name contains the sample id formatted as <sample_id>_."""
    return f"{sample_internal_id}_" in sample_fastq.name


def is_file_path_compressed_fastq(file_path: Path) -> bool:
    return file_path.name.endswith(f"{FileExtensions.FASTQ}{FileExtensions.GZIP}")


def is_lane_in_fastq_file_name(sample_fastq: Path) -> bool:
    """Validate that fastq contains lane number formatted as _L<lane_number>"""
    return bool(re.search(r"_L\d+", sample_fastq.name))


def is_valid_sample_fastq_file(sample_fastq: Path, sample_internal_id: str) -> bool:
    """
    Validate that the sample fastq file name is formatted as expected.

    Assumptions:
    1. The sample fastq file name ends with .fastq.gz
    2. The sample fastq file name contains the lane number formatted as _L<lane_number>
    3. The sample internal id is present in the parent directory name or in the file name.
    """
    sample_id_in_directory: bool = is_sample_id_in_directory_name(
        directory=sample_fastq.parent, sample_internal_id=sample_internal_id
    )
    sample_id_in_file_name: bool = is_sample_id_in_file_name(
        sample_fastq=sample_fastq, sample_internal_id=sample_internal_id
    )

    return (
        is_file_path_compressed_fastq(sample_fastq)
        and is_lane_in_fastq_file_name(sample_fastq)
        and (sample_id_in_directory or sample_id_in_file_name)
    )


def get_valid_sample_fastqs(fastq_paths: list[Path], sample_internal_id: str) -> list[Path]:
    """Return a list of valid fastq files."""
    return [
        fastq
        for fastq in fastq_paths
        if is_valid_sample_fastq_file(sample_fastq=fastq, sample_internal_id=sample_internal_id)
    ]


def get_sample_fastqs_from_flow_cell(
    flow_cell_directory: Path, sample_internal_id: str
) -> list[Path] | None:
    """Retrieve all fastq files for a specific sample in a flow cell directory."""

    # The flat output structure for NovaseqX flow cells demultiplexed with BCLConvert on hasta
    root_pattern = f"{sample_internal_id}_S*_L*_R*_*{FileExtensions.FASTQ}{FileExtensions.GZIP}"

    # The default structure for flow cells demultiplexed with bcl2fastq
    unaligned_pattern = (
        f"Unaligned*/Project_*/Sample_{sample_internal_id}"
        f"/*{FileExtensions.FASTQ}{FileExtensions.GZIP}"
    )

    # Alternative structure for bcl2fastq flow cells whose fastq files have a trailing sequence
    unaligned_alt_pattern = (
        f"Unaligned*/Project_*/Sample_{sample_internal_id}"
        f"_*/*{FileExtensions.FASTQ}{FileExtensions.GZIP}"
    )

    # The default structure for flow cells demultiplexed with bclconvert
    bcl_convert_pattern = (
        f"Unaligned*/*/{sample_internal_id}_*{FileExtensions.FASTQ}{FileExtensions.GZIP}"
    )

    # The pattern for novaseqx flow cells demultiplexed on board of the dragen
    demux_on_sequencer_pattern = (
        f"BCLConvert/fastq/{sample_internal_id}"
        f"_S*_L*_R*_*{FileExtensions.FASTQ}{FileExtensions.GZIP}"
    )

    for pattern in [
        root_pattern,
        unaligned_pattern,
        unaligned_alt_pattern,
        bcl_convert_pattern,
        demux_on_sequencer_pattern,
    ]:
        sample_fastqs: list[Path] = get_files_matching_pattern(
            directory=flow_cell_directory, pattern=pattern
        )
        valid_sample_fastqs: list[Path] = get_valid_sample_fastqs(
            fastq_paths=sample_fastqs, sample_internal_id=sample_internal_id
        )

        if valid_sample_fastqs:
            return valid_sample_fastqs


def create_delivery_file_in_flow_cell_directory(flow_cell_directory: Path) -> None:
    Path(flow_cell_directory, DemultiplexingDirsAndFiles.DELIVERY).touch()


def get_q30_threshold(sequencer_type: Sequencers) -> int:
    return FLOWCELL_Q30_THRESHOLD[sequencer_type]


def get_sample_sheet_path_from_flow_cell_dir(
    flow_cell_directory: Path,
    sample_sheet_file_name: str = DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME,
) -> Path:
    """Return the path to the sample sheet in the flow cell directory."""
    return get_file_in_directory(directory=flow_cell_directory, file_name=sample_sheet_file_name)


def add_flow_cell_name_to_fastq_file_path(fastq_file_path: Path, flow_cell_name: str) -> Path:
    """Add the flow cell name to the fastq file path if missing."""
    if is_pattern_in_file_path_name(file_path=fastq_file_path, pattern=flow_cell_name):
        LOG.debug(
            f"Flow cell name {flow_cell_name} already in {fastq_file_path}. Skipping renaming."
        )
        return fastq_file_path
    LOG.debug(f"Adding flow cell name {flow_cell_name} to {fastq_file_path}.")
    return Path(fastq_file_path.parent, f"{flow_cell_name}_{fastq_file_path.name}")


def rename_fastq_file_if_needed(fastq_file_path: Path, flow_cell_name: str) -> Path:
    """Rename the given fastq file path if the renamed fastq file path does not exist."""
    renamed_fastq_file_path: Path = add_flow_cell_name_to_fastq_file_path(
        fastq_file_path=fastq_file_path, flow_cell_name=flow_cell_name
    )
    if fastq_file_path != renamed_fastq_file_path:
        rename_file(file_path=fastq_file_path, renamed_file_path=renamed_fastq_file_path)
    return renamed_fastq_file_path


def get_sample_sheet(flow_cell: FlowCellDirectoryData) -> SampleSheet:
    """Return sample sheet associated with flowcell."""
    sample_sheet_path: Path = flow_cell.get_sample_sheet_path_hk()
    sample_sheet: SampleSheet = get_sample_sheet_from_file(sample_sheet_path)
    return sample_sheet


def get_undetermined_fastqs(lane: int, flow_cell_path: Path) -> list[Path]:
    """Get the undetermined fastq files for a specific lane on a flow cell."""
    undetermined_pattern = f"Undetermined*_L00{lane}_*{FileExtensions.FASTQ}{FileExtensions.GZIP}"
    undetermined_in_root: list[Path] = get_files_matching_pattern(
        directory=flow_cell_path,
        pattern=undetermined_pattern,
    )
    unaligned_path = Path(flow_cell_path, DemultiplexingDirsAndFiles.UNALIGNED_DIR_NAME)
    undetermined_in_unaligned: list[Path] = get_files_matching_pattern(
        directory=unaligned_path,
        pattern=undetermined_pattern,
    )
    return undetermined_in_root + undetermined_in_unaligned


def parse_manifest_file(manifest_file: Path) -> list[Path]:
    """Returns a list with the first entry of each row of the given TSV file."""
    LOG.debug(f"Parsing manifest file: {manifest_file}")
    files: list[list[str]] = read_csv(file_path=manifest_file, delimiter="\t")
    return [Path(file[0]) for file in files]


def is_file_relevant_for_demultiplexing(file: Path) -> bool:
    """Returns whether a file is relevant for demultiplexing."""
    relevant_directories = [DemultiplexingDirsAndFiles.INTER_OP, DemultiplexingDirsAndFiles.DATA]
    for relevant_directory in relevant_directories:
        if relevant_directory in file.parts:
            return True
    return False


def get_existing_manifest_file(source_directory: Path) -> Path | None:
    """Returns the first existing manifest file in the source directory."""
    manifest_file_paths = [
        Path(source_directory, DemultiplexingDirsAndFiles.ILLUMINA_FILE_MANIFEST),
        Path(source_directory, DemultiplexingDirsAndFiles.CG_FILE_MANIFEST),
    ]
    for file_path in manifest_file_paths:
        if file_path.exists():
            return file_path


def are_all_files_synced(files_at_source: list[Path], target_directory: Path) -> bool:
    """Checks if all relevant files in the source are present in the target directory."""
    for file in files_at_source:
        target_file_path = Path(target_directory, file)
        if is_file_relevant_for_demultiplexing(file) and not target_file_path.exists():
            LOG.info(f"File: {file}, has not been transferred from source to {target_directory}")
            return False
    return True


def is_syncing_complete(source_directory: Path, target_directory: Path) -> bool:
    """Returns whether all relevant files for demultiplexing have been synced from the source to the target."""
    existing_manifest_file: Path | None = get_existing_manifest_file(source_directory)

    if not existing_manifest_file:
        LOG.debug(f"{source_directory} does not contain a manifest file. Skipping.")
        return False

    files_at_source: list[Path] = parse_manifest_file(existing_manifest_file)
    return are_all_files_synced(files_at_source=files_at_source, target_directory=target_directory)


def get_flow_cell_id(flow_cell_dir_name: str) -> str:
    """Return the flow cell id from the flow cell directory name.
    Example: 230802_A00689_0857_BHGTMFDSX7 -> HGTMFDSX7
    """
    return flow_cell_dir_name.split("_")[-1][1:]


def is_manifest_file_required(flow_cell_dir: Path) -> bool:
    """Returns whether a flow cell directory needs a manifest file."""
    illumina_manifest_file = Path(flow_cell_dir, DemultiplexingDirsAndFiles.ILLUMINA_FILE_MANIFEST)
    custom_manifest_file = Path(flow_cell_dir, DemultiplexingDirsAndFiles.CG_FILE_MANIFEST)
    copy_complete_file = Path(flow_cell_dir, DemultiplexingDirsAndFiles.COPY_COMPLETE)
    sequencing_finished: bool = copy_complete_file.exists() or is_nanopore_sequencing_complete(
        flow_cell_dir
    )
    return (
        not any((illumina_manifest_file.exists(), custom_manifest_file.exists()))
        and sequencing_finished
    )


def create_manifest_file(flow_cell_dir_name: Path) -> Path:
    """Creates a tab separated file containing the paths of all files in the given
    directory and any subdirectories."""
    files_in_directory: list[list[str]] = []
    for subdir, _, files in os.walk(flow_cell_dir_name):
        subdir = Path(subdir).relative_to(flow_cell_dir_name)
        files_in_directory.extend([Path(subdir, file).as_posix()] for file in files)
    LOG.info(
        f"Writing manifest file to {Path(flow_cell_dir_name, DemultiplexingDirsAndFiles.CG_FILE_MANIFEST)}"
    )
    output_path = Path(flow_cell_dir_name, DemultiplexingDirsAndFiles.CG_FILE_MANIFEST)
    write_csv(
        content=files_in_directory,
        file_path=output_path,
        delimiter="\t",
    )
    return output_path


def is_flow_cell_sync_confirmed(target_flow_cell_dir: Path) -> bool:
    return Path(target_flow_cell_dir, DemultiplexingDirsAndFiles.COPY_COMPLETE).exists()


def get_nanopore_summary_file(flow_cell_directory: Path) -> bool | None:
    """Returns the summary file for a Nanopore run if found."""
    try:
        file = Path(next(flow_cell_directory.glob(NANOPORE_SEQUENCING_SUMMARY_PATTERN)))
    except StopIteration:
        file = None
    return file


def is_nanopore_sequencing_complete(flow_cell_directory: Path) -> bool:
    """Returns whether the sequencing is complete for a Nanopore run."""
    return bool(get_nanopore_summary_file(flow_cell_directory))



from pathlib import Path
from cg.meta.workflow.microsalt.constants import QUALITY_REPORT_FILE_NAME

from cg.meta.workflow.microsalt.quality_controller.models import (
    CaseQualityResult,
    SampleQualityResult,
)
from cg.meta.workflow.microsalt.quality_controller.report_generator import ReportGenerator


def test_generate_report_with_results(
    quality_results: list[SampleQualityResult], case_result: CaseQualityResult, tmp_path: Path
):
    # GIVEN quality results

    # GIVEN a file path to write them to
    out_file = Path(tmp_path, QUALITY_REPORT_FILE_NAME)

    # WHEN generating a report
    ReportGenerator.report(out_file=out_file, samples=quality_results, case=case_result)

    # THEN the report is created
    assert out_file.exists()

    # THEN the report is populated
    assert out_file.read_text()



"""This file tests the analyses_to_clean part of the status api"""

from datetime import datetime

from cg.constants import Workflow
from cg.store.models import CaseSample
from cg.store.store import Store


def test_analysis_included(
    analysis_store: Store, helpers, timestamp_now: datetime, timestamp_yesterday: datetime
):
    """Tests that analyses that are uploaded are returned."""

    # GIVEN an analysis that is uploaded
    analysis = helpers.add_analysis(
        analysis_store,
        started_at=timestamp_yesterday,
        uploaded_at=timestamp_yesterday,
        cleaned_at=None,
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp_yesterday)
    link: CaseSample = analysis_store.relate_sample(
        case=analysis.case, sample=sample, status="unknown"
    )
    analysis_store.session.add(link)

    # WHEN calling the analyses_to_clean
    analyses_to_clean = analysis_store.get_analyses_to_clean(before=timestamp_now)

    # THEN this analysis should be returned
    assert analysis in analyses_to_clean


def test_analysis_excluded(analysis_store: Store, helpers, timestamp_now: datetime):
    """Tests that analyses that are completed but lacks delivery report are returned."""

    # GIVEN an analysis that is not uploaded
    analysis = helpers.add_analysis(
        analysis_store, started_at=timestamp_now, uploaded_at=None, cleaned_at=None
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp_now)
    link: CaseSample = analysis_store.relate_sample(
        case=analysis.case, sample=sample, status="unknown"
    )
    analysis_store.session.add(link)

    # WHEN calling the analyses_to_clean
    analyses_to_clean = analysis_store.get_analyses_to_clean()

    # THEN this analysis should be returned
    assert analysis not in analyses_to_clean


def test_workflow_included(
    analysis_store: Store, helpers, timestamp_now: datetime, timestamp_yesterday: datetime
):
    """Tests that analyses that are included depending on workflow."""

    # GIVEN an analysis that is uploaded and workflow is specified
    workflow = Workflow.BALSAMIC
    analysis = helpers.add_analysis(
        analysis_store,
        started_at=timestamp_yesterday,
        uploaded_at=timestamp_yesterday,
        cleaned_at=None,
        workflow=workflow,
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp_yesterday)
    link: CaseSample = analysis_store.relate_sample(
        case=analysis.case, sample=sample, status="unknown"
    )
    analysis_store.session.add(link)

    # WHEN calling the analyses_to_clean specifying the used workflow
    analyses_to_clean = analysis_store.get_analyses_to_clean(
        before=timestamp_now, workflow=workflow
    )

    # THEN this analysis should be returned
    assert analysis in analyses_to_clean


def test_workflow_excluded(analysis_store: Store, helpers, timestamp_now: datetime):
    """Tests that analyses are excluded depending on pipeline."""

    # GIVEN an analysis that is uploaded

    used_workflow = Workflow.BALSAMIC
    wrong_workflow = Workflow.MIP_DNA
    analysis = helpers.add_analysis(
        analysis_store,
        started_at=timestamp_now,
        uploaded_at=timestamp_now,
        cleaned_at=None,
        workflow=used_workflow,
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp_now)
    link: CaseSample = analysis_store.relate_sample(
        case=analysis.case, sample=sample, status="unknown"
    )
    analysis_store.session.add(link)

    # WHEN calling the analyses_to_clean specifying another workflow
    analyses_to_clean = analysis_store.get_analyses_to_clean(workflow=wrong_workflow)

    # THEN this analysis should not be returned
    assert analysis not in analyses_to_clean


def test_non_cleaned_included(
    analysis_store: Store, helpers, timestamp_now: datetime, timestamp_yesterday: datetime
):
    """Tests that analyses that are included depending on cleaned_at."""

    # GIVEN an analysis that is uploaded but not cleaned
    analysis = helpers.add_analysis(
        analysis_store,
        started_at=timestamp_yesterday,
        uploaded_at=timestamp_yesterday,
        cleaned_at=None,
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp_yesterday)
    link: CaseSample = analysis_store.relate_sample(
        case=analysis.case, sample=sample, status="unknown"
    )
    analysis_store.session.add(link)

    # WHEN calling the analyses_to_clean
    analyses_to_clean = analysis_store.get_analyses_to_clean(before=timestamp_now)

    # THEN this analysis should be returned
    assert analysis in analyses_to_clean


def test_cleaned_excluded(analysis_store: Store, helpers, timestamp_now: datetime):
    """Tests that analyses are excluded depending on cleaned_at."""

    # GIVEN an analysis that is cleaned
    analysis = helpers.add_analysis(
        analysis_store,
        started_at=timestamp_now,
        uploaded_at=timestamp_now,
        cleaned_at=timestamp_now,
    )
    sample = helpers.add_sample(analysis_store, delivered_at=timestamp_now)
    link: CaseSample = analysis_store.relate_sample(
        case=analysis.case, sample=sample, status="unknown"
    )
    analysis_store.session.add(link)

    # WHEN calling the analyses_to_clean
    analyses_to_clean = analysis_store.get_analyses_to_clean()

    # THEN this analysis should not be returned
    assert analysis not in analyses_to_clean



from cg.server.dto.orders.orders_response import Order, OrdersResponse
from cg.store.models import Order as DatabaseOrder


def parse_order(order: DatabaseOrder) -> Order:
    return Order(
        customer_id=order.customer.internal_id,
        ticket_id=order.ticket_id,
        order_date=str(order.order_date.date()),
        order_id=order.id,
        workflow=order.workflow,
    )


def create_orders_response(database_orders: list[DatabaseOrder]) -> OrdersResponse:
    parsed_database_orders: list[Order] = [parse_order(order) for order in database_orders]
    return OrdersResponse(orders=parsed_database_orders)


def create_order_response(order: DatabaseOrder) -> Order:
    return parse_order(order)



"""Module for Taxprofiler Analysis API."""

import logging
from pathlib import Path
from typing import Any

from cg.constants import Workflow
from cg.constants.sequencing import SequencingPlatform
from cg.io.json import read_json
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import MetricsBase, MultiqcDataJson
from cg.models.fastq import FastqFileMeta
from cg.models.taxprofiler.taxprofiler import (
    TaxprofilerParameters,
    TaxprofilerSampleSheetEntry,
)
from cg.store.models import Case, Sample

LOG = logging.getLogger(__name__)


class TaxprofilerAnalysisAPI(NfAnalysisAPI):
    """Handles communication between Taxprofiler processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.TAXPROFILER,
    ):
        super().__init__(config=config, workflow=workflow)
        self.root_dir: str = config.taxprofiler.root
        self.nfcore_workflow_path: str = config.taxprofiler.pipeline_path
        self.conda_env: str = config.taxprofiler.conda_env
        self.conda_binary: str = config.taxprofiler.conda_binary
        self.profile: str = config.taxprofiler.profile
        self.revision: str = config.taxprofiler.revision
        self.hostremoval_reference: Path = Path(config.taxprofiler.hostremoval_reference)
        self.databases: Path = Path(config.taxprofiler.databases)
        self.tower_binary_path: str = config.tower_binary_path
        self.tower_workflow: str = config.taxprofiler.tower_pipeline
        self.account: str = config.taxprofiler.slurm.account
        self.email: str = config.taxprofiler.slurm.mail_user
        self.nextflow_binary_path: str = config.taxprofiler.binary_path
        self.compute_env_base: str = config.taxprofiler.compute_env

    def get_sample_sheet_content_per_sample(
        self, sample: Sample, instrument_platform: SequencingPlatform.ILLUMINA, fasta: str = ""
    ) -> list[list[str]]:
        """Get sample sheet content per sample."""
        sample_name: str = sample.name
        sample_metadata: list[FastqFileMeta] = self.gather_file_metadata_for_sample(sample)
        fastq_forward_read_paths: list[str] = self.extract_read_files(
            metadata=sample_metadata, forward_read=True
        )
        fastq_reverse_read_paths: list[str] = self.extract_read_files(
            metadata=sample_metadata, reverse_read=True
        )
        sample_sheet_entry = TaxprofilerSampleSheetEntry(
            name=sample_name,
            run_accession=sample_name,
            instrument_platform=instrument_platform,
            fastq_forward_read_paths=fastq_forward_read_paths,
            fastq_reverse_read_paths=fastq_reverse_read_paths,
            fasta=fasta,
        )
        return sample_sheet_entry.reformat_sample_content()

    def get_sample_sheet_content(
        self,
        case_id: str,
        instrument_platform: SequencingPlatform.ILLUMINA,
        fasta: str = "",
    ) -> list[list[Any]]:
        """Write sample sheet for Taxprofiler analysis in case folder."""
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        sample_sheet_content = []
        LOG.info(f"Samples linked to case {case_id}: {len(case.links)}")
        LOG.debug("Getting sample sheet information")
        for link in case.links:
            sample_sheet_content.extend(
                self.get_sample_sheet_content_per_sample(
                    sample=link.sample, instrument_platform=instrument_platform, fasta=fasta
                )
            )
        return sample_sheet_content

    def get_workflow_parameters(self, case_id: str) -> TaxprofilerParameters:
        """Return Taxprofiler parameters."""
        LOG.debug("Getting parameters information")
        return TaxprofilerParameters(
            cluster_options=f"--qos={self.get_slurm_qos_for_case(case_id=case_id)}",
            sample_sheet_path=self.get_sample_sheet_path(case_id=case_id),
            outdir=self.get_case_path(case_id=case_id),
            databases=self.databases,
            hostremoval_reference=self.hostremoval_reference,
            priority=self.account,
        )

    def config_case(
        self,
        case_id: str,
        instrument_platform: SequencingPlatform.ILLUMINA,
        dry_run: bool,
        fasta: str = "",
    ) -> None:
        """Create sample sheet file and parameters file for Taxprofiler analysis."""
        self.create_case_directory(case_id=case_id, dry_run=dry_run)
        sample_sheet_content: list[list[Any]] = self.get_sample_sheet_content(
            case_id=case_id,
            instrument_platform=instrument_platform,
            fasta=fasta,
        )
        workflow_parameters: TaxprofilerParameters = self.get_workflow_parameters(case_id=case_id)
        if dry_run:
            LOG.info("Dry run: Config files will not be written")
            return
        self.write_sample_sheet(
            content=sample_sheet_content,
            file_path=self.get_sample_sheet_path(case_id=case_id),
            header=TaxprofilerSampleSheetEntry.headers(),
        )
        self.write_params_file(case_id=case_id, workflow_parameters=workflow_parameters.dict())

    def get_multiqc_json_metrics(self, case_id: str) -> list[MetricsBase]:
        """Return a list of the metrics specified in a MultiQC json file for the case samples."""
        multiqc_json: list[dict] = MultiqcDataJson(
            **read_json(file_path=self.get_multiqc_json_path(case_id=case_id))
        ).report_general_stats_data
        samples: list[Sample] = self.status_db.get_samples_by_case_id(case_id=case_id)
        metrics_list: list[MetricsBase] = []
        for sample in samples:
            sample_id: str = sample.internal_id
            metrics_values: dict = self.parse_multiqc_json_for_sample(
                sample_name=sample.name, multiqc_json=multiqc_json
            )
            metric_base_list: list = self.get_metric_base_list(
                sample_id=sample_id, metrics_values=metrics_values
            )
            metrics_list.extend(metric_base_list)
        return metrics_list

    @staticmethod
    def parse_multiqc_json_for_sample(sample_name: str, multiqc_json: list[dict]) -> dict:
        """Parse a multiqc_data.json and returns a dictionary with metric name and metric values for each sample."""
        metrics_values: dict = {}
        for stat_dict in multiqc_json:
            for sample_key, sample_values in stat_dict.items():
                if sample_key == f"{sample_name}_{sample_name}":
                    LOG.info(f"Key: {sample_key}, Values: {sample_values}")
                    metrics_values.update(sample_values)

        return metrics_values



"""Module for defining invoice models."""

from typing import Any

from pydantic.v1 import BaseModel

from cg.constants.priority import PriorityTerms
from cg.constants.sequencing import RecordType


class InvoiceContact(BaseModel):
    """Class for collection contact information used in the invoice."""

    name: str
    email: str
    customer_name: str
    reference: str
    address: str


class InvoiceApplication(BaseModel):
    """Class to collect Application information used in the invoice."""

    version: str
    tag: str
    discounted_price: int
    percent_kth: int


class InvoiceInfo(BaseModel):
    """Class to collect invoice information used in the invoice report."""

    name: str
    id: str
    lims_id: str | None
    application_tag: str
    project: str
    date: Any | None
    price: int
    priority: PriorityTerms
    price_kth: int | None
    total_price: int | None


class InvoiceReport(BaseModel):
    """Class that collects information used to create the invoice Excel sheet."""

    cost_center: str
    project_number: str | None
    customer_id: str
    customer_name: str
    agreement: str | None
    invoice_id: int
    contact: dict
    records: list
    pooled_samples: Any | None
    record_type: RecordType



from enum import StrEnum

from pydantic import BaseModel, BeforeValidator, ConfigDict, constr
from typing_extensions import Annotated

from cg.constants import DataDelivery, Workflow
from cg.models.orders.validators.sample_base_validators import snake_case
from cg.store.models import Application, Case, Customer, Pool, Sample


class ControlEnum(StrEnum):
    not_control = ""
    positive = "positive"
    negative = "negative"


class SexEnum(StrEnum):
    male = "male"
    female = "female"
    unknown = "unknown"


class PriorityEnum(StrEnum):
    research = "research"
    standard = "standard"
    priority = "priority"
    express = "express"
    clinical_trials = "clinical_trials"


class ContainerEnum(StrEnum):
    no_container = "No container"
    plate = "96 well plate"
    tube = "Tube"


class StatusEnum(StrEnum):
    affected = "affected"
    unaffected = "unaffected"
    unknown = "unknown"


NAME_PATTERN = r"^[A-Za-z0-9-]*$"


class OrderSample(BaseModel):
    model_config = ConfigDict(
        arbitrary_types_allowed=True, populate_by_name=True, coerce_numbers_to_str=True
    )
    age_at_sampling: str | None = None
    application: constr(max_length=Application.tag.property.columns[0].type.length)
    capture_kit: str | None = None
    collection_date: str | None = None
    comment: constr(max_length=Sample.comment.property.columns[0].type.length) | None = None
    concentration: float | None = None
    concentration_ng_ul: float | None = None
    concentration_sample: float | None = None
    container: ContainerEnum | None = None
    container_name: str | None = None
    control: str | None = None
    customer: constr(max_length=Customer.internal_id.property.columns[0].type.length) | None = None
    custom_index: str | None = None
    data_analysis: Workflow
    data_delivery: DataDelivery
    elution_buffer: str | None = None
    extraction_method: str | None = None
    family_name: (
        constr(
            pattern=NAME_PATTERN,
            min_length=2,
            max_length=Case.name.property.columns[0].type.length,
        )
        | None
    ) = None
    father: (
        constr(pattern=NAME_PATTERN, max_length=Sample.name.property.columns[0].type.length) | None
    ) = None
    formalin_fixation_time: int | None = None
    index: str | None = None
    index_number: str | None = None
    index_sequence: str | None = None
    internal_id: constr(max_length=Sample.internal_id.property.columns[0].type.length) | None = None
    lab_code: str | None = None
    mother: (
        constr(pattern=NAME_PATTERN, max_length=Sample.name.property.columns[0].type.length) | None
    ) = None
    name: constr(
        pattern=NAME_PATTERN,
        min_length=2,
        max_length=Sample.name.property.columns[0].type.length,
    )
    organism: str | None = None
    organism_other: str | None = None
    original_lab: str | None = None
    original_lab_address: str | None = None
    phenotype_groups: list[str] | None = None
    phenotype_terms: list[str] | None = None
    pool: constr(max_length=Pool.name.property.columns[0].type.length) | None = None
    post_formalin_fixation_time: int | None = None
    pre_processing_method: str | None = None
    priority: Annotated[PriorityEnum, BeforeValidator(snake_case)] = PriorityEnum.standard
    primer: str | None = None
    quantity: int | None = None
    reagent_label: str | None = None
    reference_genome: (
        constr(max_length=Sample.reference_genome.property.columns[0].type.length) | None
    ) = None
    region: str | None = None
    region_code: str | None = None
    require_qc_ok: bool = False
    rml_plate_name: str | None = None
    selection_criteria: str | None = None
    sex: SexEnum = SexEnum.unknown
    source: str | None = None
    status: StatusEnum = StatusEnum.unknown
    subject_id: (
        constr(pattern=NAME_PATTERN, max_length=Sample.subject_id.property.columns[0].type.length)
        | None
    ) = None
    tissue_block_size: str | None = None
    tumour: bool = False
    tumour_purity: int | None = None
    verified_organism: bool | None = None
    volume: str | None = None
    well_position: str | None = None
    well_position_rml: str | None = None



"""Tests for rsync API"""

import logging
import shutil
from pathlib import Path

import pytest

from cg.constants.priority import SlurmAccount, SlurmQos
from cg.exc import CgError
from cg.meta.rsync import RsyncAPI
from cg.store.models import Case
from cg.store.store import Store
from tests.meta.deliver.conftest import all_samples_in_inbox, dummy_file_name
from tests.store.conftest import case_obj


def test_get_source_and_destination_paths(
    mutant_case: Case, rsync_api: RsyncAPI, ticket_id: str, mocker
):
    """Test generating the source path before rsync."""

    # GIVEN a valid Sars-cov-2 case
    case = mutant_case

    # GIVEN file exists
    mocker.patch.object(RsyncAPI, "get_all_cases_from_ticket")
    RsyncAPI.get_all_cases_from_ticket.return_value = [case]

    # WHEN the source path is created
    source_and_destination_paths = rsync_api.get_source_and_destination_paths(
        ticket=ticket_id, customer_internal_id=case.customer.internal_id
    )

    # THEN the source path ends with a customer id, followed by "inbox" and a ticket_id id
    assert (
        source_and_destination_paths["delivery_source_path"]
        .as_posix()
        .endswith(f"/cust000/inbox/{ticket_id}")
    )
    # THEN the destination path is in the format server.name.se:/path/cust_id/path/ticket_id/
    assert (
        source_and_destination_paths["rsync_destination_path"].as_posix()
        == "server.name.se:/some/cust000/inbox"
    )


def test_set_log_dir(rsync_api: RsyncAPI, ticket_id: str, caplog):
    """Test function to set log dir for path."""

    caplog.set_level(logging.INFO)

    # GIVEN an RsyncAPI, with its base path as its log dir
    base_path: Path = rsync_api.log_dir

    # WHEN setting the log directory
    rsync_api.set_log_dir(folder_prefix=ticket_id)

    # THEN the log dir should set to a new path, different from the base path
    assert base_path.as_posix() != rsync_api.log_dir.as_posix()
    assert "Setting log dir to:" in caplog.text


def test_make_log_dir(rsync_api: RsyncAPI, ticket_id: str, caplog):
    """Test generating the directory for logging."""
    caplog.set_level(logging.INFO)

    # WHEN the log directory is created
    rsync_api.set_log_dir(folder_prefix=ticket_id)
    rsync_api.create_log_dir(dry_run=True)

    # THEN the path is not created since it is a dry run
    assert "Would have created path" in caplog.text

    # THEN the created path is
    assert str(rsync_api.log_dir).startswith(f"/another/path/{ticket_id}")


def test_run_rsync_on_slurm(
    microsalt_case: Case, rsync_api: RsyncAPI, ticket_id: str, caplog, mocker, helpers
):
    """Test for running rsync using SLURM."""
    caplog.set_level(logging.INFO)

    # GIVEN a valid microsalt case
    case: Case = microsalt_case

    # GIVEN paths needed to run rsync
    mocker.patch.object(RsyncAPI, "get_source_and_destination_paths")
    RsyncAPI.get_source_and_destination_paths.return_value = {
        "delivery_source_path": Path("/path/to/source"),
        "rsync_destination_path": Path("/path/to/destination"),
    }

    mocker.patch.object(RsyncAPI, "get_all_cases_from_ticket")
    RsyncAPI.get_all_cases_from_ticket.return_value = [case]

    # WHEN the destination path is created
    sbatch_number: int = rsync_api.run_rsync_on_slurm(ticket=ticket_id, dry_run=True)

    # THEN check that SARS-COV-2 analysis is not delivered
    assert "Delivering report for SARS-COV-2 analysis" not in caplog.text

    # THEN check that an integer was returned as sbatch number
    assert isinstance(sbatch_number, int)


def test_run_rsync_on_slurm_no_cases(rsync_api: RsyncAPI, ticket_id: str, caplog, mocker, helpers):
    """Test for running rsync using SLURM when there are no cases on the ticket."""
    caplog.set_level(logging.INFO)

    # GIVEN ticket without any cases
    mocker.patch.object(RsyncAPI, "get_all_cases_from_ticket")
    RsyncAPI.get_all_cases_from_ticket.return_value = None

    # WHEN the job is submitted
    with pytest.raises(CgError):
        rsync_api.run_rsync_on_slurm(ticket=ticket_id, dry_run=True)

        # THEN check that error is raised based on no cases being present
        assert "Could not find any cases for ticket" in caplog.text


def test_get_folders_to_deliver(
    analysis_family: dict, analysis_store_trio, rsync_api: RsyncAPI, case_id: str
):
    """Tests the ability for the rsync api to get case and sample names."""
    # GIVEN a case

    # WHEN the function gets the folders
    folder_list: list[str] = rsync_api.get_folders_to_deliver(
        case_id=case_id, sample_files_present=True, case_files_present=True
    )

    # THEN it the list should contain the case name and all the samples
    assert folder_list == [
        analysis_family["samples"][0]["name"],
        analysis_family["samples"][1]["name"],
        analysis_family["samples"][2]["name"],
        analysis_family["name"],
    ]


def test_concatenate_rsync_commands(
    analysis_family: dict, analysis_store_trio, project_dir, customer_id, ticket_id: str
):
    """Tests the function to concatenate rsync commands for transferring multiple files."""
    # GIVEN a list with a case and a sample name
    folder_list: list[str] = [analysis_family["name"], analysis_family["samples"][0]["name"]]
    source_and_destination_paths = {
        "delivery_source_path": project_dir / customer_id / ticket_id,
        "rsync_destination_path": project_dir / customer_id,
    }
    # WHEN then commands are generated
    commands: str = RsyncAPI.concatenate_rsync_commands(
        folder_list=folder_list,
        source_and_destination_paths=source_and_destination_paths,
        ticket=ticket_id,
    )
    # THEN the correct folder should be added to the source path
    assert (
        " ".join(
            [
                str(source_and_destination_paths["delivery_source_path"] / analysis_family["name"]),
                str(source_and_destination_paths["delivery_source_path"]),
            ]
        )
        in commands
    )
    assert (
        " ".join(
            [
                str(
                    source_and_destination_paths["delivery_source_path"]
                    / analysis_family["samples"][0]["name"]
                ),
                str(source_and_destination_paths["delivery_source_path"]),
            ]
        )
        in commands
    )


def test_slurm_rsync_single_case(
    all_samples_in_inbox: Path,
    case: Case,
    destination_path: Path,
    rsync_api: RsyncAPI,
    caplog,
    mocker,
    ticket_id: str,
):
    """Test for running rsync on a single case using SLURM."""
    caplog.set_level(logging.INFO)

    # GIVEN paths needed to run rsync
    mocker.patch.object(RsyncAPI, "get_source_and_destination_paths")
    RsyncAPI.get_source_and_destination_paths.return_value = {
        "delivery_source_path": all_samples_in_inbox,
        "rsync_destination_path": destination_path,
    }

    mocker.patch.object(Store, "get_latest_ticket_from_case")
    Store.get_latest_ticket_from_case.return_value = ticket_id

    # WHEN the destination path is created
    sbatch_number: int
    is_complete_delivery: bool
    is_complete_delivery, sbatch_number = rsync_api.slurm_rsync_single_case(
        case=case,
        case_files_present=True,
        dry_run=True,
        sample_files_present=True,
    )

    # THEN check that an integer was returned as sbatch number and the delivery should be complete
    assert isinstance(sbatch_number, int)
    assert is_complete_delivery


def test_slurm_rsync_single_case_missing_file(
    all_samples_in_inbox: Path,
    case: Case,
    destination_path: Path,
    rsync_api: RsyncAPI,
    caplog,
    mocker,
    ticket_id: str,
):
    """Test for running rsync on a single case with a missing file using SLURM."""
    caplog.set_level(logging.INFO)

    # GIVEN a valid mip case and sample folder missing
    shutil.rmtree(Path(all_samples_in_inbox, case.links[0].sample.name))

    # GIVEN paths needed to run rsync
    mocker.patch.object(RsyncAPI, "get_source_and_destination_paths")
    RsyncAPI.get_source_and_destination_paths.return_value = {
        "delivery_source_path": all_samples_in_inbox,
        "rsync_destination_path": destination_path,
    }

    mocker.patch.object(Store, "get_latest_ticket_from_case")
    Store.get_latest_ticket_from_case.return_value = ticket_id

    # WHEN the destination path is created
    sbatch_number: int
    is_complete_delivery: bool
    is_complete_delivery, sbatch_number = rsync_api.slurm_rsync_single_case(
        case=case,
        case_files_present=True,
        dry_run=True,
        sample_files_present=True,
    )

    # THEN check that an integer was returned as sbatch number
    assert isinstance(sbatch_number, int)
    assert not is_complete_delivery


def test_slurm_quality_of_service_production(rsync_api: RsyncAPI):
    # GIVEN an RsyncAPI instance

    # WHEN the account of the api is set to production
    rsync_api.account = SlurmAccount.PRODUCTION

    # THEN the quality of service should be set to HIGH
    assert rsync_api.slurm_quality_of_service == SlurmQos.HIGH


def test_slurm_quality_of_service_other(rsync_api: RsyncAPI):
    # GIVEN an RsyncAPI instance

    # WHEN the account of the api is set to development
    rsync_api.account = SlurmAccount.DEVELOPMENT

    # THEN the quality of service should be set to LOW
    assert rsync_api.slurm_quality_of_service == SlurmQos.LOW



import pytest

from cg.meta.backup.backup import BackupAPI
from cg.meta.encryption.encryption import EncryptionAPI
from cg.meta.tar.tar import TarAPI
from cg.models.cg_config import CGConfig


@pytest.fixture
def backup_context(cg_context: CGConfig) -> CGConfig:
    cg_context.meta_apis["backup_api"] = BackupAPI(
        encryption_api=EncryptionAPI(binary_path=cg_context.encryption.binary_path),
        pdc_archiving_directory=cg_context.backup.pdc_archiving_directory,
        status=cg_context.status_db,
        tar_api=TarAPI(binary_path=cg_context.tar.binary_path),
        pdc_api=cg_context.pdc_api,
        flow_cells_dir=cg_context.flow_cells_dir,
    )
    return cg_context



"""Tests for the metrics-deliver cli command."""

import logging

from _pytest.logging import LogCaptureFixture
from click.testing import CliRunner

from cg.cli.workflow.rnafusion.base import metrics_deliver
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig


def test_without_samples(
    cli_runner: CliRunner,
    rnafusion_context: CGConfig,
    caplog: LogCaptureFixture,
    no_sample_case_id: str,
):
    """Test command with case_id and no samples."""
    caplog.set_level(logging.ERROR)
    # GIVEN case-id

    # WHEN dry running with dry specified
    result = cli_runner.invoke(
        metrics_deliver, [no_sample_case_id, "--dry-run"], obj=rnafusion_context
    )

    # THEN command should NOT execute successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN warning should be printed that no samples exist in case
    assert "no samples" in caplog.text


def test_metrics_deliver(
    cli_runner: CliRunner,
    rnafusion_context: CGConfig,
    rnafusion_mock_analysis_finish,
    caplog: LogCaptureFixture,
    rnafusion_case_id: str,
):
    """Test command with case_id and analysis_finish which should execute successfully."""
    caplog.set_level(logging.INFO)
    # GIVEN case-id

    # WHEN dry running with dry specified
    result = cli_runner.invoke(metrics_deliver, [rnafusion_case_id], obj=rnafusion_context)

    # THEN command should execute successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN rnafusion and case_id should be found in command string
    assert "Writing metrics deliverables file to" in caplog.text



import logging
from datetime import datetime

import petname

from cg.constants import DataDelivery, FlowCellStatus, Priority, Workflow
from cg.constants.archiving import PDC_ARCHIVE_LOCATION
from cg.models.orders.order import OrderIn
from cg.store.base import BaseHandler
from cg.store.database import get_session
from cg.store.models import (
    Analysis,
    Application,
    ApplicationLimitations,
    ApplicationVersion,
    Bed,
    BedVersion,
    Case,
    CaseSample,
    Collaboration,
    Customer,
    Delivery,
    Flowcell,
    Invoice,
    Order,
    Organism,
    Panel,
    Pool,
    Sample,
    SampleLaneSequencingMetrics,
    User,
)

LOG = logging.getLogger(__name__)


class CreateHandler(BaseHandler):
    """Methods related to adding new data to the store."""

    def generate_readable_sample_id(self) -> str:
        """Generates a petname as sample internal id for development purposes. Not used in normal production flow."""
        while True:
            random_id: str = petname.Generate(3, separator="")
            if not self.get_sample_by_internal_id(random_id):
                return random_id

    def generate_readable_case_id(self) -> str:
        while True:
            random_id: str = petname.Generate(2, separator="", letters=10)
            if not self.get_case_by_internal_id(random_id):
                return random_id

    def add_customer(
        self,
        internal_id: str,
        name: str,
        invoice_address: str,
        invoice_reference: str,
        data_archive_location: str = PDC_ARCHIVE_LOCATION,
        scout_access: bool = False,
        is_clinical: bool = False,
        *args,
        **kwargs,
    ) -> Customer:
        """Build a new customer record."""

        return Customer(
            internal_id=internal_id,
            name=name,
            scout_access=scout_access,
            invoice_address=invoice_address,
            invoice_reference=invoice_reference,
            data_archive_location=data_archive_location,
            is_clinical=is_clinical,
            **kwargs,
        )

    def add_collaboration(self, internal_id: str, name: str, **kwargs) -> Collaboration:
        """Build a new customer group record."""

        return Collaboration(internal_id=internal_id, name=name, **kwargs)

    def add_user(self, customer: Customer, email: str, name: str, is_admin: bool = False) -> User:
        """Build a new user record."""

        new_user = User(name=name, email=email, is_admin=is_admin)
        new_user.customers.append(customer)
        return new_user

    def add_application(
        self,
        tag: str,
        prep_category: str,
        description: str,
        percent_kth: int,
        percent_reads_guaranteed: int,
        is_accredited: bool = False,
        min_sequencing_depth: int = 0,
        **kwargs,
    ) -> Application:
        """Build a new application record."""

        return Application(
            tag=tag,
            prep_category=prep_category,
            description=description,
            is_accredited=is_accredited,
            min_sequencing_depth=min_sequencing_depth,
            percent_kth=percent_kth,
            percent_reads_guaranteed=percent_reads_guaranteed,
            **kwargs,
        )

    def add_application_version(
        self,
        application: Application,
        version: int,
        valid_from: datetime,
        prices: dict,
        **kwargs,
    ) -> ApplicationVersion:
        """Build a new application version record."""

        new_record = ApplicationVersion(version=version, valid_from=valid_from, **kwargs)
        for price_key in [
            Priority.standard.name,
            Priority.priority.name,
            Priority.express.name,
            Priority.research.name,
        ]:
            setattr(new_record, f"price_{price_key}", prices[price_key])
        new_record.application = application
        return new_record

    @staticmethod
    def add_application_limitation(
        application: Application,
        workflow: str,
        limitations: str,
        comment: str = "Dummy comment",
        created_at: datetime = datetime.now(),
        updated_at: datetime = datetime.now(),
        **kwargs,
    ) -> ApplicationLimitations:
        """Build a new application limitations record."""
        return ApplicationLimitations(
            application=application,
            pipeline=workflow,
            limitations=limitations,
            comment=comment,
            created_at=created_at,
            updated_at=updated_at,
            **kwargs,
        )

    def add_bed(self, name: str) -> Bed:
        """Build a new bed record."""
        return Bed(name=name)

    def add_bed_version(self, bed: Bed, version: int, filename: str, shortname: str) -> BedVersion:
        """Build a new bed version record."""
        bed_version: BedVersion = BedVersion(
            version=version, filename=filename, shortname=shortname
        )
        bed_version.bed = bed
        return bed_version

    def add_sample(
        self,
        name: str,
        sex: str,
        comment: str = None,
        control: str = None,
        downsampled_to: int = None,
        internal_id: str = None,
        order: str = None,
        ordered: datetime = None,
        priority: Priority = None,
        received: datetime = None,
        original_ticket: str = None,
        tumour: bool = False,
        **kwargs,
    ) -> Sample:
        """Build a new Sample record."""

        internal_id = internal_id or self.generate_readable_sample_id()
        priority = priority or (Priority.research if downsampled_to else Priority.standard)
        return Sample(
            comment=comment,
            control=control,
            downsampled_to=downsampled_to,
            internal_id=internal_id,
            is_tumour=tumour,
            name=name,
            order=order,
            ordered_at=ordered or datetime.now(),
            original_ticket=original_ticket,
            priority=priority,
            received_at=received,
            sex=sex,
            **kwargs,
        )

    def add_case(
        self,
        data_analysis: Workflow,
        data_delivery: DataDelivery,
        name: str,
        ticket: str,
        panels: list[str] | None = None,
        cohorts: list[str] | None = None,
        priority: Priority | None = Priority.standard,
        synopsis: str | None = None,
        customer_id: int | None = None,
    ) -> Case:
        """Build a new Case record."""

        internal_id: str = self.generate_readable_case_id()
        return Case(
            cohorts=cohorts,
            data_analysis=str(data_analysis),
            data_delivery=str(data_delivery),
            internal_id=internal_id,
            name=name,
            panels=panels,
            priority=priority,
            synopsis=synopsis,
            tickets=ticket,
            customer_id=customer_id,
        )

    def relate_sample(
        self,
        case: Case,
        sample: Sample,
        status: str,
        mother: Sample = None,
        father: Sample = None,
    ) -> CaseSample:
        """Relate a sample record to a family record."""

        new_record: CaseSample = CaseSample(status=status)
        new_record.case = case
        new_record.sample = sample
        new_record.mother = mother
        new_record.father = father
        return new_record

    def add_flow_cell(
        self,
        flow_cell_name: str,
        sequencer_name: str,
        sequencer_type: str,
        date: datetime,
        flow_cell_status: str | None = FlowCellStatus.ON_DISK,
        has_backup: bool | None = False,
    ) -> Flowcell:
        """Build a new Flowcell record."""
        return Flowcell(
            name=flow_cell_name,
            sequencer_name=sequencer_name,
            sequencer_type=sequencer_type,
            sequenced_at=date,
            status=flow_cell_status,
            has_backup=has_backup,
        )

    def add_analysis(
        self,
        workflow: Workflow,
        version: str = None,
        completed_at: datetime = None,
        primary: bool = False,
        uploaded: datetime = None,
        started_at: datetime = None,
        **kwargs,
    ) -> Analysis:
        """Build a new Analysis record."""
        return Analysis(
            pipeline=workflow,
            pipeline_version=version,
            completed_at=completed_at,
            is_primary=primary,
            uploaded_at=uploaded,
            started_at=started_at,
            **kwargs,
        )

    def add_panel(
        self,
        customer: Customer,
        name: str,
        abbrev: str,
        version: float,
        date: datetime = None,
        genes: int = None,
    ) -> Panel:
        """Build a new panel record."""

        new_record = Panel(
            name=name, abbrev=abbrev, current_version=version, date=date, gene_count=genes
        )
        new_record.customer = customer
        return new_record

    def add_pool(
        self,
        customer: Customer,
        name: str,
        order: str,
        ordered: datetime,
        application_version: ApplicationVersion,
        ticket: str = None,
        comment: str = None,
        received_at: datetime = None,
        invoice_id: int = None,
        no_invoice: bool = None,
        delivered_at: datetime = None,
    ) -> Pool:
        """Build a new Pool record."""

        new_record: Pool = Pool(
            name=name,
            ordered_at=ordered or datetime.now(),
            order=order,
            ticket=ticket,
            received_at=received_at,
            comment=comment,
            delivered_at=delivered_at,
            invoice_id=invoice_id,
            no_invoice=no_invoice,
        )
        new_record.customer = customer
        new_record.application_version = application_version
        return new_record

    def add_delivery(
        self,
        destination: str,
        sample: Sample = None,
        pool: Pool = None,
        comment: str = None,
    ) -> Delivery:
        """Build a new Delivery record."""

        if not any([sample, pool]):
            raise ValueError("you have to provide a sample or a pool")
        new_record: Delivery = Delivery(destination=destination, comment=comment)
        new_record.sample = sample
        new_record.pool = pool
        return new_record

    def add_invoice(
        self,
        customer: Customer,
        samples: list[Sample] = None,
        microbial_samples: list[Sample] = None,
        pools: list[Pool] = None,
        comment: str = None,
        discount: int = 0,
        record_type: str = None,
        invoiced_at: datetime | None = None,
    ):
        """Build a new Invoice record."""

        new_id = self.new_invoice_id()
        new_invoice: Invoice = Invoice(
            comment=comment,
            discount=discount,
            id=new_id,
            record_type=record_type,
            invoiced_at=invoiced_at,
        )
        new_invoice.customer = customer
        for sample in samples or []:
            new_invoice.samples.append(sample)
        for microbial_sample in microbial_samples or []:
            new_invoice.samples.append(microbial_sample)
        for pool in pools or []:
            new_invoice.pools.append(pool)
        return new_invoice

    def add_organism(
        self,
        internal_id: str,
        name: str,
        reference_genome: str = None,
        verified: bool = False,
        **kwargs,
    ) -> Organism:
        """Build a new Organism record."""
        return Organism(
            internal_id=internal_id,
            name=name,
            reference_genome=reference_genome,
            verified=verified,
            **kwargs,
        )

    def add_sample_lane_sequencing_metrics(
        self, flow_cell_name: str, sample_internal_id: str, **kwargs
    ) -> SampleLaneSequencingMetrics:
        """Add a new SampleLaneSequencingMetrics record."""
        return SampleLaneSequencingMetrics(
            flow_cell_name=flow_cell_name,
            sample_internal_id=sample_internal_id,
            **kwargs,
        )

    def add_order(self, order_data: OrderIn):
        customer: Customer = self.get_customer_by_internal_id(order_data.customer)
        order = Order(
            customer_id=customer.id,
            ticket_id=order_data.ticket,
            workflow=order_data.order_type,
        )
        session = get_session()
        session.add(order)
        session.commit()
        return order
